#!/usr/bin/perl

$| = 1;

# ddrcombine.pl   ("DDRescue logfile COMBINEr")
#
# This is a Perl program which reads two or more GNU ddrescue 1.x "log files"
# (sector list files), representing different sections of a hard disk drive,
# and combines them.
#
# The assumption is that a ddrescue log file was split into sections by
# ddrsplit.pl, which is a utility to create multile small log files from one
# large log file.  Each small log file represents a section (sector range) of
# the drive.  In each small log file, the other sectors (that is, the sectors
# outside the sector range represented by that small log file), are marked "+"
# to prevent ddrescue from reading them.
#
# This program recombines those small log files back into one big log file.
#
# (Why would you want to do such a thing?  For performance.  ddrescue bogs down
# and gets slower and slower as its log file grows.  By splitting the log file
# into sections, and processing each section separately, it can be made to go
# much faster.)
#
# This program is similar to ddrlogand.pl, in that both do something akin to
# a logical .AND. function on ddrescue log files.  But they are not quite the
# same.  For one thing, ddrcombine.pl can process any number of input log
# files, but ddrlogand.pl can only handle two.  But, more importantly, there
# is a subtle difference in how the sector statuses are calculated:
#
#                                               ddrlogand      ddrcombine
#               log_input1      log_input2        result         result
#               ----------      ----------        ------         ------
#  if sector is     -               +                -              -
#  if sector is     -               -                -            error*
#  if sector is     -               /                -            error*
#  if sector is     /               +                /              /
#  if sector is     /               -                /            error*
#  if sector is     /               /                /            error*
#  if sector is     +               +                +              +
#  if sector is     +               -                /              -
#  if sector is     +               /                /              /
#
# *Note #1:  the four error cases are not detected by this program.
# Instead, such sectors are classified in the output file as they were
# classified in log_input1 (the first input file).
#
# Note #2:  absent the error cases, the order of the input files should
# not affect the result.  However, the program runs fastest when the input
# files are in their "natural" order -- i.e., with the input log file
# representing the beginning of the disk specified first, and the input
# log file representing the end of the disk specified last on the command
# line.
#
# Note #3:  you can, if you wish, specify one of the input file names as
# the name of the output file (since the program doesn't write the output
# file until it finishes reading all the input files).  So, this command:
#
#   perl ddrcombine.pl comb.log x.log-01 x.log-02 x.log-03 x.log-04 x.log-05
#
# is equivalent to this series of commands:
#
#   perl ddrcombine.pl comb.log x.log-01 x.log-02
#   perl ddrcombine.pl comb.log comb.log x.log-03
#   perl ddrcombine.pl comb.log comb.log x.log-04 x.log-05
#
# Under Linux and BSD, the shell handles wild-card expansion, so you can do:
#
#   perl ddrcombine.pl comb.log x.log-*
#
# If, for some reason, you wanted to run this Perl program under Windows,
# you could use a DOS 'for loop' consruct for wild-card expansion:
#
#   copy x.log-01 comb.log
#   for %f in (x.log-*) do perl ddcombine.pl comb.log comb.log %f
#



# BTW, here's a hint.  Open a console and do:
#   tail -s 3 -f /var/log/messages
# to watch the messages logged during a ddrescue attempt.


# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#
#
# v.1 - 26-Oct-2005 - first version.  -DAB
# v.2 - 17-Sep-2009 - accomodate the status indicator in the ddrescue logfile.  -DAB

#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "8, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


$verbose = $summarize = 0;


$perlver = "3 or earlier";
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
}


@sv_ARGV = @ARGV;  # save raw parameter list, for display to the user

# display command-line parameters (just once)
$args_shown = 0;
sub showargs {
   if (!$args_shown) {
      print join( " ", "perl $0", @sv_ARGV ) . "\n";  # echo command-line arguments
      $args_shown = 1;
   }
}


# Handle input parameters and open files:

# check for '-v' and/or '-s'
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-[vs]/)) {
   foreach $ch (split(//,$ARGV[0])) {
      if ('v' eq $ch) {
         $verbose++;
      } elsif ('s' eq $ch) {
         $summarize++;
      }
   }
   shift @ARGV;
}

if ($#ARGV < 2) {
   die "ddrcombine.pl   (\"DDRescue logfile COMBINEr\")   v.$ver_ordinal_and_date\n" .
       "\n" .
       "Usage:\n" .
       "   perl -w $0 outputfile inputfile1 inputfile2 ...\n" .
       "\n" .
       "or, add '-v' for verbose output (or '-vv' for very verbose):\n" .
       "   perl -w $0 -v outputfile inputfiles...\n" .
       "\n" .
       "or, to print a summary at the end (via ddrsummarize.pl), add \"-s\":\n" .
       "   perl -w $0 -s outputfile inputfiles...\n" .
       "\n" .
       "*** Note the odd parameter order (output file first)! ***\n" .
       "\n" .
       "Also, note that '-v' can be repeated for very verbose output.\n" .
       "\n" .
       "More detailed instructions are included in the source code comments,\n" .
       "at the beginning of $0.\n";
}

&showargs;

if ($verbose) {
   print "You are using Perl version $perlver\n";
}

$log_output = shift @ARGV;
$log_input1 = shift @ARGV;
open( LOG_INPUT1, "<$log_input1") || die "ERROR: Could not read $log_input1, $!\n";


######################################################################
# Here's a sample of what a ddrescue log file used to look like:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.1
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
#
# Here's a sample of what a ddrescue log file looks like now; line 3 is
# the status line; lines 2 & 3 were introduced in ddrescue 1.6:
#
#    # Rescue Logfile. Created by GNU ddrescue version 1.9
#    # current_pos  current_status
#    #    0x00000000     ?
#    #      pos        size  status
#    0x00000000  0x0004BE00  +
#    0x0004BE00  0x00000200  -
#    0x0004C000  0x33078E00  +
#    0x330C4E00  0x00001000  -
#    0x330C5E00  0x8ACDF200  +
#
######################################################################


# input is an integer represented as a string (probably hexadecimal preceeded by '0x'),
# result is the integer divided by 512.
sub divby512 {
   local($dividend) = @_;
   # print "dividend=$dividend\n";
   local($result) = 0;
   if ($dividend =~ /^0x/) {
      # print "tricky way!\n";
      # tricky divide by 512: drop the last 2 hexidecimal digits, and divide by 2 (which is more robust for very large integers)
      $result = substr($dividend,0,length($dividend)-2);
      $result = int(eval($result) / 2);  # gotta use eval() because it starts with 0x
   } else {
      # print "conventional way,\n";
      $result = int((0+$dividend) / 512);
   }
   # print "result=$result\n";
   return $result;
}


# Internal representation of the sector list is a bit tricky.
# The problem is that there are soooo many sectors.  A 300 GB
# disk drive has 600 million sectors.  Even at one character
# per sector, that's too big to represent in memory as a simple
# array or string.  So we have to represent it like it is in
# the sector list file, as a list of ranges.  But that means
# that we can't use a simple array or hash lookup to check the
# status of a particular sector; we'll have to call a function
# to do a search.



if ($verbose) {
   print "reading   '$log_input1'\n";
}

# First, read log_input1 (the first ddrescue sector list input file)
$numranges = 0;
$lastline = '';
$highest_from_inp1 = 0;
while (<LOG_INPUT1>) {
   if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
      $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
      # Parse a ddrescue log file line:
      @tmp = split( /\s+/, $_ );
      # for ($i=0; $i<=$#tmp; $i++) { print "dbg: tmp[$i]='" . $tmp[$i] . "'\n"; }
      if ((1 == $#tmp) && ($tmp[1] =~ /^[\?\*\/\-\+]$/)) {
         # the new status line (introduced in ddrescue 1.6)
         next;
      }
      ($pos_byt, $len_byt, $status) = @tmp;
      if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
         print "ERROR: unrecognized line in '$log_input1':\n$_\n";
         next;
      }
      if (('-' ne $status) && ('+' ne $status)) {
         # ignore all but '-' (error) regions and '+' (successful) regions
         $lastline = $_;
         next;
      }
      $lastline = '';
      if ($len_byt eq '-1') {
         print "ERROR: length=-1 for status = '$status' in '$log_input1':\n$_\n";
         next;
      }

      # Now convert byte offsets and counts to LBA sector number & sector count:
      $pos_sect = &divby512($pos_byt);
      $len_sect = &divby512($len_byt);

      # print "dbg: Bad sect no. = $pos_sect, count = $len_sect, '$status'\n";
      # for ($i=0; $i<$len_sect; $i++) {
      #   $originally_bad_sectors{$pos_sect} = 1;
      #   $num_originally_bad_sectors++;
      #   $pos_sect++;
      # }

      $offsets[$numranges] = $pos_sect;
      $lengths[$numranges] = $len_sect;
      $stati[$numranges] = $status;  # '-' or '+'
      $numranges++;
      $highest_from_inp1 = $pos_sect + $len_sect;
      if ($status eq '') { print "gak1\n";}  # dbg
   }
}
close LOG_INPUT1;


$count_rescued1 = $count_failed1 = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued1 += $len;
   } elsif ('-' eq $status) {
      $count_failed1 += $len;
   }
}


# Format a ddrescue logfile line.  Input parameters are in sectors (512 bytes),
# but the result is displayed in bytes.
sub displayrange {
   local( $ofs, $len, $status ) = @_;
   if (!defined $status) {
      $status = 'undef';  # should never happen
   }
   local( $result ) = sprintf( "0x%06X00  0x%06X00  %s", (2*$ofs), (2*$len), $status );
   return $result;
}


# this subroutine assumes that $log_input2 is open on LOG_INPUT2
sub read_another_input_file {
   local( $pos_byt, $len_byt, $status, @tmp, $pos_sect, $len_sect );
   $numranges2 = 0;
   $lastline2 = '';
   while (<LOG_INPUT2>) {
      if (($_ =~ /\d/) && ($_ !~ /^\#/)) {
         $_ =~ s/\s*$//;  # strip trailing blanks, tabs, cr, lf, etc.
         # Parse a ddrescue log file line:
         @tmp = split( /\s+/, $_ );
         if ((1 == $#tmp) && ($tmp[1] =~ /^[\?\*\/\-\+]$/)) {
            # the new status line (introduced in ddrescue 1.6)
            next;
         }
         ($pos_byt, $len_byt, $status) = @tmp;
         if (($#tmp != 2) || ($pos_byt !~ /\d/) || ($len_byt !~ /\d/)) {
            print "ERROR: unrecognized line in '$log_input2':\n$_\n";
            next;
         }
         if ('+' eq $status) {
            # ignore all but '-' (error) regions and '/' (uncopied) regions
            $lastline2 = $_;
            next;
         }
         $lastline2 = '';

         # Now convert byte offsets and counts to LBA sector number & sector count:
         $pos_sect = &divby512($pos_byt);
         if ($len_byt eq '-1') {
            $len_sect = $highest_from_inp1;
         } else {
            $len_sect = &divby512($len_byt);
         }
         $offsets2[$numranges2] = $pos_sect;
         $lengths2[$numranges2] = $len_sect;
         $stati2[$numranges2] = $status;  # '-' or '/'
         $numranges2++;
         # print "dbg: $log_input2.$numranges2: " . &displayrange( $pos_sect, $len_sect, $status ) . "\n";
      }
   }
}


# sub tst {
#    print "dbg: last = " . &displayrange( $offsets[$numranges-1], $lengths[$numranges-1], $stati[$numranges-1] ) . ", ofs+len=" . ($offsets[$numranges-1] + $lengths[$numranges-1]) . "\n";
# }


# Merge the two sector lists.
# this subroutine assumes that &read_another_input_file has loaded an input
# file into @offsets2/@lengths2/@stati2/$numranges2.  That information is used
# to modify @offsets/@lengths/@stati/$numranges.
sub merge_another_input_file {
   local( $1, $j, $ofs, $len, $status, $ofs2, $len2, $status2 );
   # Step through the older sector list:
   for ($i=0; $i<$numranges; $i++) {
      # the only regions we care about are the '+' regions.
      if ('+' eq $stati[$i]) {
         $status = '+';
         $ofs = $offsets[$i];
         $len = $lengths[$i];
         # was any part of this region not '+' in the newer sector list?
         for ($j=0; $j<$numranges2; $j++) {
            $ofs2 = $offsets2[$j];
            $len2 = $lengths2[$j];
            if (($ofs2+$len2) <= $ofs) {
               next;
            }
            $status2 = $stati2[$j];
            if ($ofs2 >= ($ofs+$len)) {
               if ($verbose >= 3) {
                  print "dbg: i=$i, no overlap of:\n" .
                        "  Old: " . &displayrange($ofs,$len,$status) . "\n";
                  if ($j > 0) {
                     print "  new: " . &displayrange($offsets2[$j-1],$lengths2[$j-1],$stati2[$j-1]) . "\n";
                  }
                  print "  New: " . &displayrange($ofs2,$len2,$status2) . "\n";
               }
               last;  # since the regions are in order, we're done
            }
            if ($verbose >= 2) {
               print "dbg: i=$i, j=$j, found an overlap.\n" .
                     "  input1: " . &displayrange($ofs,$len,$status) . "\n" .
                     "  input2: " . &displayrange($ofs2,$len2,$status2) . "\n";
            }
            # They overlap!  That means that we need to mark part or all of
            # the '+' range with the status from the newer file ('-' or '/').
            # There are four cases:
            # 1) The '-' or '/' range covers the entire '+' region
            # 2) The '-' or '/' range covers the beginning of the '+' region
            # 3) The '-' or '/' range covers the end of the '+' region
            # 4) The '-' or '/' range covers the middle, but not the beginning or end
            if (($ofs2 <= $ofs) && (($ofs2+$len2) >= ($ofs+$len))) {
               # Case #1: The '-' or '/' range covers the entire '+' region
               $stati[$i] = $status2;
               if ($verbose >= 2) {
                  print "  Entire '+' range changed to '$status2'.\n";
               }
            } elsif ($ofs2 <= $ofs) {
               # Case #2: The '-' or '/' range covers the beginning of the '+' region
               if ($ofs2 < 0) { print "gak3\n";}  # dbg
               if ($len2 < -1) { print "gak4\n";}  # dbg
               if (!defined $status2) { printf "gak2:  0x%09X00  0x%09x00  ''\n",$ofs2*2,$len2*2;  $status2='x';}  # dbg
               if ('/' eq $status2) {
                  $offsets[$i] = $ofs2+$len2;
                  $lengths[$i] = $ofs+$len - ($ofs2+$len2);
                  if ($verbose >= 2) {
                     print "  Beginning of '+' range changed to '/'.\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  new: " . &displayrange( $offsets[$i], $lengths[$i], '-' ) . "\n";
                  }
               } elsif ('-' eq $status2) {
                  $lengths[$i] = $ofs2+$len2-$ofs;
                  splice( @offsets, $i+1, 0, $ofs2+$len2 );
                  splice( @lengths, $i+1, 0, $len-($ofs2+$len2-$ofs) );
                  splice( @stati, $i+1, 0, '+' );
                  $stati[$i] = $status2;
                  $numranges++;
                  if ($verbose >= 2) {
                     print "  Beginning of '+' range changed to '-'.  Split old range:\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  pt1: " . &displayrange( $offsets[$i], $lengths[$i], '-' ) . "\n" .
                           "  pt2: " . &displayrange( $offsets[$i+1], $lengths[$i+1], '+' ) . "\n";
                  }
               } else {
                  print "ERR: gak2\n";
               }
            } elsif (($ofs2+$len2) >= ($ofs+$len)) {
               # 3) The '-' or '/' range covers the end of the '+' region
               # There are two sub-cases.
               # To change it to '/' is simple: since we don't represent '/' regions in
               # the list, we can just adjust (reduce) the sector range for the '+'
               # region, and we're done.  But to change it to '-' requires splitting
               # the region:
               $lengths[$i] = $ofs2-$ofs;
               if ('/' eq $status2) {
                  if ($verbose >= 2) {
                     print "  End of '+' range changed to '$status2'.\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  new: " . &displayrange( $offsets[$i], $lengths[$i], '+' ) . "\n";
                  }
               } else {
                  splice( @offsets, $i+1, 0, $ofs2 );
                  splice( @lengths, $i+1, 0, $len-($ofs2-$ofs) );
                  splice( @stati, $i+1, 0, $status2 );
                  $numranges++;
                  if ($verbose >= 2) {
                     print "  End of '+' range changed to '-'.  Split old range:\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  pt1: " . &displayrange( $offsets[$i], $lengths[$i], '+' ) . "\n" .
                           "  pt2: " . &displayrange( $offsets[$i+1], $lengths[$i+1], '-' ) . "\n";
                  }
               }
            } else {
               # Case #4 is the hardest, because we need to split the '+' region
               # into two parts if the new file has it as '/', and into three parts
               # if the new file as it as '-'
               if ($ofs2 < 0) { print "gak5\n";}  # dbg
               if ($len2 < -1) { print "gak6\n";}  # dbg
               if (! defined $status2) { printf "gak7:  0x%09X00  0x%09x00  'undef'\n",$ofs2*2,$len2*2;}  # dbg
               $lengths[$i] = $ofs2 - $ofs;
               if ('/' eq $status2) {
                  splice( @offsets, $i+1, 0, $ofs2+$len2 );
                  splice( @lengths, $i+1, 0, ($ofs+$len)-($ofs2+$len2) );
                  splice( @stati, $i+1, 0, '+' );
                  $numranges++;
                  if ($verbose >= 2) {
                     print "  Middle of '+' range changed to '/'.  Split old range:\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  pt1: " . &displayrange( $offsets[$i], $lengths[$i], '+' ) . "\n" .
                           "  pt2: " . &displayrange( $offsets[$i+1], $lengths[$i+1], '+' ) . "\n";
                  }
               } else {
                  splice( @offsets, $i+1, 0, $ofs2, $ofs2+$len2 );
                  splice( @lengths, $i+1, 0, $len2, ($ofs+$len)-($ofs2+$len2) );
                  splice( @stati, $i+1, 0, $status2, '+' );
                  $numranges += 2;
                  if ($verbose >= 2) {
                     print "  Middle of '+' range changed to '-'.  Split old range into 3 parts:\n";
                     print "    i=$i, len = $len -> " . $lengths[$i] . "\n";
                     print "  pt1: " . &displayrange( $offsets[$i], $lengths[$i], '+' ) . "\n" .
                           "  pt2: " . &displayrange( $offsets[$i+1], $lengths[$i+1], $status2 ) . "\n" .
                           "  pt3: " . &displayrange( $offsets[$i+2], $lengths[$i+2], '+' ) . "\n";
                  }
               }
            }

            # in case we changed the region, update $ofs, $len & status
            $ofs = $offsets[$i];
            $len = $lengths[$i];
            $status = $stati[$i];
            if ($verbose >= 2) {
               print "dbg:  ofs/len/st = " . &displayrange( $ofs, $len, $status ) . "\n";
            }
            # &tst; # dbg
            if ('+' ne $status) {
               last;  # increment $i (go on to the next '+' region)
            }
         }
      }
   }
}


sub process_another_input_file {
   $log_input2 = shift(@_);  # intentionally global, so read_another_input_file() and merge_another_input_file() can see it
   if (!open( LOG_INPUT2, "<$log_input2")) {
      print "ERROR: Could not read $log_input2, $!\n";
   } else {
      if ($verbose) {
         print "combining '$log_input2'\n";
      }
      &read_another_input_file;
      close LOG_INPUT2;
      # for ($x = 0; $x <= $numranges2-1; $x++) {
      #    print "dbg: $log_input2/$x : " . &displayrange( $offsets2[$x], $lengths2[$x], $stati2[$x] ) . "\n";
      # }
      # &tst; # dbg
      &merge_another_input_file;
      undef @offsets2;
      undef @lengths2;
      undef @stati2;
      $numranges2 = 0;
   }
}


for (;;) {
   if ($#ARGV < 0) {
      last;
   }
   $log_input2 = shift @ARGV;
   if ($log_input2 =~ /^-v/) {
      $verbose++;  # '-v' flag, for verbose processing of subsequent input files
   } else {
      &process_another_input_file( $log_input2 );
   }
}


$count_rescued = $count_failed = 0;
for ($i=0; $i<$numranges; $i++) {
   $status = $stati[$i];
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   if ('+' eq $status) {
      $count_rescued += $len;
   } elsif ('-' eq $status) {
      $count_failed += $len;
   }
}
if ($verbose >= 2) {
   print "dbg: '+'  $count_rescued sectors were rescued.\n";
   print "dbg: '-'  $count_failed sectors failed.\n";
}


# now merge any adjacent lines with the same status
for ($i = 1; $i < $numranges; $i++) {
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   $status = $stati[$i];
   if (($status eq $stati[$i-1]) && ($ofs == ($offsets[$i-1]+$lengths[$i-1]))) {
      if ($verbose >= 2) {
         print "merge:\n" .
               "  pt1: " . &displayrange( $offsets[$i-1], $lengths[$i-1], $stati[$i-1] ) . "\n" .
               "  pt2: " . &displayrange( $offsets[$i], $lengths[$i], $stati[$i] ) . "\n";
      }
      $lengths[$i-1] += $lengths[$i];
      $offsets[$i] = $ofs+$len;
      $lengths[$i] = 0;
      $stati[$i] = '/';
      if ($verbose >= 2) {
         print "  new: " . &displayrange( $offsets[$i-1], $lengths[$i-1], $stati[$i-1] ) . "\n";
      }
   }
}


# Now spit out the list:

open( LOG_OUTPUT, ">$log_output" ) || die "ERROR: Could not create $log_output, $!\n";
binmode LOG_OUTPUT;  # write output in binary mode, to force LF-only line termination even under DOS/Windows
if ($verbose) {
   print "writing   '$log_output\n";
}
print LOG_OUTPUT "# Rescue Logfile. Created by GNU ddrescue version 1.9-burtonhack2\n" .
      "# current_pos  current_status\n" .
      "0x00000000     ?\n" .
      "#      pos        size  status\n";
$prev_ofs_plus_len = 0;
for ($i = 0; $i < $numranges; $i++) {
   $ofs = $offsets[$i];
   $len = $lengths[$i];
   $status = $stati[$i];
   if ('/' eq $status) {
      next;
   }
   if ($ofs < $prev_ofs_plus_len) {
      print "ERROR: overlapping regions:\n  " . &displayrange($ofs,$len,$status) . "\n";;
   } elsif ($ofs > $prev_ofs_plus_len) {
      $gaplen = $ofs - $prev_ofs_plus_len;
      print LOG_OUTPUT &displayrange($prev_ofs_plus_len,$gaplen,'/') . "\n";
   }
   print LOG_OUTPUT &displayrange($ofs,$len,$status) . "\n";
   $prev_ofs_plus_len = $ofs + $len;
}

if ($lastline ne '') {
   ($pos_byt) = split( /\s+/, $lastline );
   $ofs = &divby512($pos_byt);
   if ($ofs < $prev_ofs_plus_len) {
       printf "ERROR: overlapping region at end of %s:\n%s\n", $log_input1, $lastline;
   } elsif ($ofs > $prev_ofs_plus_len) {
       $gaplen = $ofs - $prev_ofs_plus_len;
       print LOG_OUTPUT &displayrange($prev_ofs_plus_len,$gaplen,'/') . "\n";
   }
   print LOG_OUTPUT "$lastline\n";
}
close LOG_OUTPUT;

# '-s' parameter:
if ($summarize) {
   $ddrsummarize = 'ddrsummarize.pl';
   if ($0 =~ /^(.*[\/\\\:])([^\/\\\:]+)$/) {
      $ddrsummarize = $1 . $ddrsummarize;
   }
   if ($verbose) {
      print "Run '$ddrsummarize'...\n";
   }
   $loaded_via_REQUIRE = 0;  $loaded_via_REQUIRE++;  # define a flag so ddrsummarize.pl knows we called it (two stmts instead of one to avoid perl -w warning)
   @ARGV = ($log_output);  # parameter list to ddrsummarize
   require $ddrsummarize;
}

__END__

